from similar_documents.cmd_utils import (
    files_to_texts,
    similar_vectors_top_k,
    assign_top_k,
)
from similar_documents.similar import tfidf_vectorize
from similar_documents.tokenizer import japanese

from pathlib import Path
import os

fixtures_path = Path(os.path.dirname(__file__)).joinpath("fixtures")


def test_integration():
    files = [str(path) for path in fixtures_path.glob("*")]
    texts = files_to_texts(files)
    vectors = tfidf_vectorize(texts, tokenizer=japanese)
    top_ks = similar_vectors_top_k(vectors, k=2)
    assigned = assign_top_k(files, top_ks)
    assert len(list(assigned.keys())) == 5
    assert len(list(assigned.values())[0]) == 2  # size of k



from similar_documents.tokenizer import japanese


def test_japanese():
    assert japanese("こんにちは、かわいい犬ですね。") == ["こんにちは", "かわいい", "犬"]
    assert japanese("猫と犬") == ["猫", "犬"]


def test_japanese_numeric():
    assert japanese("こんにちは、100か千ですね。") == ["こんにちは"]



from similar_documents.similar import tfidf_vectorize, top_k
from similar_documents.tokenizer import japanese


def test_similar():
    texts = ["hello world!", "hello similar!", "hi, hello", "similar", "hi"]
    vectors = tfidf_vectorize(texts)
    results = list(top_k(vectors[4], vectors))  # 4 is hi
    assert results[0][0] == 2  # 2 is 'hi hello'
    results = list(top_k(vectors[3], vectors))
    assert results[0][0] == 1


def test_similar_japanese():
    texts = ["こんにちは、かわいい犬ですね。", "犬とライオン", "猫とライオン"]
    vectors = tfidf_vectorize(texts, tokenizer=japanese)
    results = list(top_k(vectors[0], vectors))
    assert results[0][0] == 1



from similar_documents.text_converter import markdown, text, html

from pathlib import Path
import os

fixtures_path = Path(os.path.dirname(__file__)).joinpath("fixtures")


def _read_fixture(filename: str):
    target = fixtures_path.joinpath(filename)
    return target.read_text(encoding="utf-8")


def test_markdown():
    fixture = _read_fixture("markdown.md")
    assert markdown(fixture) == "Hello markdown"


def test_markdown_with_yaml():
    fixture = _read_fixture("markdown_yaml.md")
    assert markdown(fixture) == "Title\nHello markdown with yaml"


def test_markdown_yaml_empty_title():
    fixture = _read_fixture("markdown_yaml_empty_title.md")
    assert markdown(fixture) == "empty title"


def test_html():
    fixture = _read_fixture("1.html")
    assert html(fixture) == "Hello html"


def test_text():
    fixture = _read_fixture("plain.txt")
    assert text(fixture) == "plain text<br />"



from __future__ import annotations
from typing import Any, cast


def _createJapaneseTokenizer(target_features=["名詞", "形容詞", "感動詞"], ignore_numeric=True):
    import fugashi

    Tagger = cast(Any, fugashi).Tagger
    tagger = Tagger("-Owakati")

    def tokenizer(text: str) -> list[str]:
        tokens = []
        for word in tagger(text):
            if word.feature[0] in target_features:
                if not (ignore_numeric and word.feature[1] == "数詞"):
                    tokens.append(word.surface)
        return tokens

    return tokenizer


japanese = _createJapaneseTokenizer()


from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import Optional, cast, Any

import plac as _plac

from . import assign_top_k, files_to_texts, similar_vectors_top_k, tfidf_vectorize
from . import tokenizer as _tokenizer

plac = cast(Any, _plac)


@plac.opt("output_file", help="Optional: write output file (default STDOUT)", type=Path)
@plac.opt("top_k", help="Number of similar documents", abbrev="k", type=int)
@plac.opt(
    "tokenizer",
    help="tokenizer [japanese] (default None(space splitting))",
    choices=["japanese"],
)
@plac.opt("encoding", help="file encoding(default utf-8)")
@plac.flg("debug", "Show debug messages")
def cli(
    output_file: Optional[Path] = None,
    top_k=5,
    debug=False,
    tokenizer=None,
    encoding="utf-8",
    *_documents: str,
):
    """
    usage: $ similar-documents -o result.json -k 5 -t japanese *.md
           $ similar-documents -h
    """
    if len(_documents) == 0:
        print(cli.__doc__, file=sys.stderr)
        exit(1)

    documents = list(_documents)

    if tokenizer:
        if tokenizer and tokenizer == "japanese":
            tokenizer = _tokenizer.japanese
        else:
            raise Exception(f"tokenizer:{tokenizer} is unknown.")
    if debug:
        print(f"files to texts {len(documents)} documents", file=sys.stderr)
    texts = files_to_texts(documents, encoding=encoding)
    if debug:
        print(f"calc tfidf...", file=sys.stderr)
    vectors = tfidf_vectorize(texts, tokenizer=tokenizer)
    if debug:
        print(f"calc similarity...", file=sys.stderr)
    top_ks = similar_vectors_top_k(vectors, k=top_k)
    if debug:
        print(f"assign similarity score", file=sys.stderr)
    assigned = assign_top_k(documents, top_ks)
    json_body = json.dumps(assigned)
    if output_file:
        if debug:
            print(f"write result to file {str(output_file)}", file=sys.stderr)
        output_file.open("w", encoding=encoding).write(json_body)
    else:
        print(json_body)



from __future__ import annotations
from typing import Callable, Optional, cast

import numpy as np
from scipy.sparse.csr import csr_matrix
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity


def tfidf_vectorize(
    texts: list[str], tokenizer: Optional[Callable[[str], list[str]]] = None
) -> csr_matrix:
    vectorizer_kargs = {
        "tokenizer": tokenizer,
    }
    if tokenizer:
        vectorizer_kargs["token_pattern"] = None
    vectorizer = TfidfVectorizer(**vectorizer_kargs)
    return cast(csr_matrix, vectorizer.fit_transform(texts))


def top_k(target: csr_matrix, vectors: csr_matrix, k=5) -> list[tuple[int, float]]:
    scores = cosine_similarity(target, vectors)[0]
    sort_indexes = scores.argsort()[::-1]
    top_indexes = sort_indexes[: 1 + k]
    top_scores = scores[top_indexes]
    if abs(top_scores[0] - 1.0) < 0.001:
        # delete top-1
        top_scores = np.delete(top_scores, 0)
        top_indexes = np.delete(top_indexes, 0)
    elif top_scores.shape[0] > k:
        top_scores = np.delete(top_scores, k)
        top_indexes = np.delete(top_indexes, k)
    return list(zip(top_indexes, top_scores))


from . import text_converter, tokenizer

from .cmd_utils import files_to_texts, similar_vectors_top_k, assign_top_k
from .similar import tfidf_vectorize
from .cli import cli

version = "0.1.4"

__all__ = [
    "cli",
    "files_to_texts",
    "similar_vectors_top_k",
    "assign_top_k",
    "tfidf_vectorize",
    "text_converter",
    "tokenizer",
    "version",
]



from __future__ import annotations
from typing import Callable

from markdown import Markdown
from bs4 import BeautifulSoup

TextConverter = Callable[[str], str]


def _createMarkdownToText():
    md = Markdown(extensions=["full_yaml_metadata"])

    def md2text(source: str) -> str:
        body = md.convert(source)
        meta = md.Meta  # type: ignore # XXX...
        text = html(str(body))
        if type(meta) is dict:
            for key in meta.keys():
                if key.upper() == "TITLE" and type(meta[key]) is str:
                    text = f"{meta[key]}\n{text}"
                    break

        return text

    return md2text


markdown = _createMarkdownToText()


def html(source: str) -> str:
    return "".join(
        BeautifulSoup(source, features="html.parser").findAll(text=True)
    ).strip()


def text(source: str) -> str:
    return source.strip()


from __future__ import annotations

import itertools
import multiprocessing as mp
from pathlib import Path
from typing import Optional

from scipy.sparse.csr import csr_matrix

from . import similar, text_converter

_ext_converter_mappings = (
    (("md", "mdx", "markdown"), text_converter.markdown),
    (("html", "htm"), text_converter.html),
)


def _detect_converter(ext: str):
    ext = ext.lower().replace(".", "")
    for (exts, converter) in _ext_converter_mappings:
        if ext in exts:
            return converter
    return text_converter.text


def file_to_text(
    file: str,
    converter: Optional[text_converter.TextConverter] = None,
    encoding="utf-8",
) -> str:
    path = Path(file)
    source = path.read_text(encoding=encoding)
    if converter:
        converted = converter(source)
    else:
        converted = _detect_converter(path.suffix)(source)
    return converted


def _file_to_text_tuple(t: tuple(str, Optional[text_converter.TextConverter], str)):  # type: ignore
    return file_to_text(t[0], converter=t[1], encoding=t[2])


def files_to_texts(
    files: list[str],
    converter: Optional[text_converter.TextConverter] = None,
    encoding="utf-8",
) -> list[str]:
    pool = mp.Pool(mp.cpu_count())
    texts: list[str] = pool.map(
        _file_to_text_tuple,
        zip(files, itertools.cycle([converter]), itertools.cycle([encoding])),
    )
    return texts


def _top_k_tuple(t: tuple(csr_matrix, csr_matrix, int)):  # type: ignore
    return similar.top_k(*t)


def similar_vectors_top_k(vectors: csr_matrix, k=5):
    pool = mp.Pool(mp.cpu_count())
    return pool.map(
        _top_k_tuple, zip(vectors, itertools.cycle([vectors]), itertools.cycle([k]))
    )


def assign_top_k(files: list[str], top_ks: list[list[tuple[int, float]]]):
    file_dict: dict[str, list[tuple[str, float]]] = {}
    for (file, top_k) in zip(files, top_ks):
        scores = [(files[index], score) for (index, score) in top_k]
        file_dict[file] = scores
    return file_dict
